@@ -3,12 +3,12 @@ context("PCMSim")
 
 library(PCMBase)
 
-if(PCMBaseIsADevRelease(numVersionComponents = 3)) {
+if(PCMBaseIsADevRelease()) {
 
 
-  load("testobjects.RData")
+  list2env(PCMBaseTestObjects, globalenv())
 
-  set.seed(1)
+  set.seed(1, kind = "Mersenne-Twister", normal.kind = "Inversion")
 
   # generate traits
   traits.a.1 <- PCMSim(tree.a, model.a.1, 0, verbose=TRUE)
@@ -37,8 +37,4 @@ if(PCMBaseIsADevRelease(numVersionComponents = 3)) {
     }
   )
 
-  # save testobjects:
-  #
-  # save(a.H, a.Sigma_x, a.Sigmae_x, b.H, b.Sigma_x, b.Sigmae_x, model_MixedGaussian_ab, model_MixedGaussian_ab_globalSigmae_x, model.a.1, model.a.1.Omitted_X0, model.a.123, model.a.123.Omitted_X0, model.a.123.Omitted_X0__bSigmae_x, model.a.123.Omitted_X0__Omitted_Sigmae_x, model.a.2, model.a.3, model.ab.123, model.ab.123.bSigmae_x, model.b.123, model.b.123.Omitted_X0, model.b.123.Omitted_X0__Omitted_Sigmae_x, tree.a, tree.ab, Theta, a.Theta, a.X0, b.Theta, b.X0, H, N, Sigma_x, Sigmae_x, traits.a.1, traits.a.123, traits.a.2, traits.a.3, traits.ab.123, model.a.1, file = "tests/testthat/testobjects.RData")
-
 }
